/*
 * gc - Gram's Commander: A Norton-Commanderish UNIX utility version 2.2
 *
 * YAGH! (tm) (Yet Another Gramish Hack)
 *
 * Please contribute suggestions/code/bug reports, etc. If you find
 * this program very useful, you are welcome to send a small contribution
 * to the author, at P.O.Box 15525, Vlaeberg, 8018. This will encourage the
 * development of enhancements and other utilities you always wish you had!
 *
 * This program may be freely copied and distributed, provided this message
 * and the title screen are retained and all modifications are noted
 * and accredited at the end of this message.
 *
 * Written by Graham Wheeler, Aztec Information Managment
 * 	gram@aim1.aztec.co.za
 *
 * Use the accompanying Makefile to compile and install.
 *
 * Version History
 * ---------------
 * v1.0	  10 Nov 1992		Initial release
 * v2.0	  15 Nov 1992		Largely rewritten to be table-driven
 * v2.1    4 Dec 1992		Support for non-ANSI C compilers contributed
 *					by George Sipe
 *				Support for old ioctls TCGETA and TCSETAW
 *					for those with no POSIX-style
 *					tcgetattr and tcsetattr.
 * v2.2				Termio ioctls removed.Hopefully much
 *					more portable. Command mode
 *					has been taken out. A few
 *					bugs fixed. Placeholder replacement
 *					now done in a preprocessing phase
 *					before parsing commands.
 *
 * DISCLAIMER: The author is not responsible for any loss, damage,
 *	riots, earthquakes, catastrophes, marriages, annulments,
 *	etc that arise from the	use of this program. Sorry...
 *
 */

#include <sys/types.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>

#ifdef ULTRIX
#	include <cursesX.h>
#	include <sys/stat.h>
#	define MAXNAMLEN _DIRH_MAXNAMLEN
#else
#	include <curses.h>
#endif

#include <fcntl.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <malloc.h>

#include "gc.h"

static char *iCmdNames[] = {
	"paint",	"quit",		"help",		
	"read",		"wait",		"search",	"clr",
	"swap",		"up",		"down",		"pgup",
	"pgdn",		"home",		"end",		"inv",
	"invall",	"beep",		"cd",	
	NULL
};

static internalCmd_t iCmdTbl[] = {
	i_paint,	i_quit,		i_help,	
	i_read,		i_wait,		i_search,	i_clr,
	i_swap,		i_up,		i_down,		i_pgup,
	i_pgdn,		i_home,		i_end,		i_inv,
	i_invall,	i_beep,		i_cd,
	NULL
};

#ifdef DEBUG
FILE *debug;
#endif

static cmdInfo_t cmdTable[MAX_CMDS];
static keyInfo_t keyTable[138];
static char delimChar = ':';
static char metaChar = '@';

static int	l;	/* Current active panel	*/
static int	f;	/* Current file index	*/
static char *fname;	/* Current file name	*/
static int ch = 0;	/* Current key press	*/
static int	inCurses = 0;
static int	noDirSelect = 0;

static fInfo_t fInfo[2][MAXARGS];

static char	paths[2][1024],
		scrap[1024+256],
		srchTxt[256],
		*oldSelNames[MAXARGS],
		*tempName,
		cmdline[1024];

static int	numfiles[2]={0,0},
		fIndex[2][MAXARGS],
		start, end,
		fileNameLen,
		sortStyle[2],
		secSortStyle[2],
		highlight[2],
		startLn[2],
		selCnt[2],
		reverseSort[2],
		secReverseSort[2],
		ignoreDirChange,
		inMultiLoop,
		numCmds = 0,
		searching = 0,
		totallen;
static long	selSize[2];
static WINDOW	*panels[2];
static WINDOW	*panelTop[2];
static WINDOW	*statusWin[2];
static WINDOW	*statusTop[2];
static WINDOW	*cmdWin;

/***************************************************************
		SET OPERATIONS
***************************************************************/

static int
#if	__STDC__
inSet(set_t *s, int v) {
#else
inSet(s, v)
	set_t (*s);
	int v;
{
#endif /* __STDC__ */
	int i = v/32, j = v%32;
	return (((*s)[i] &(1l << j)) != 0l);
}

static int
#if	__STDC__
emptySet(set_t *s) {
#else
emptySet(s)
	set_t (*s);
{
#endif /* __STDC__ */
	int i;
	for (i=0;i<SETSIZE;i++) if ((*s)[i]) return 0;
	return 1;
}

static void
#if	__STDC__
add2set(set_t *s, int v) {
#else
add2set(s, v)
	set_t (*s);
	int v;
{
#endif /* __STDC__ */
	int i = v/32, j = v%32;
	(*s)[i] |= (1l << j);
}

static void
#if	__STDC__
clearSet(set_t *s) {
#else
clearSet(s)
	set_t (*s);
{
#endif /* __STDC__ */
	int i;
	for (i=0;i<SETSIZE;i++) (*s)[i] = 0l;
}

static void
#if	__STDC__
addRange2set(set_t *s, int sv, int ev) {
#else
addRange2set(s, sv, ev)
	set_t (*s);
	int sv;
	int ev;
{
#endif /* __STDC__ */
	while (sv<=ev) add2set(s,sv++);
}

/***********************************************
	KEY DEFINITION DUMPING
***********************************************/

static char pb[3]={'\0','\0','\0'};

#if	__STDC__
static char *pr(int ch) {
#else
static char *pr(ch)
	int ch;
{
#endif /* __STDC__ */
	
	if (ch<32) { pb[0]='^'; pb[1] = ch+64; }
	else if (ch>=128) { pb[0]='F'; pb[1]=ch-128; }
	else if (ch=='F' || ch=='\\' || ch=='^' || ch==' ') { pb[0]='\\'; pb[1]=ch; }
	else { pb[0]=ch; pb[1] = '\0'; }
	return pb;
}

static void
#if	__STDC__
dumpKeys(void) {
#else
dumpKeys()
{
#endif /* __STDC__ */
	int i, j; 
	char *p;
	FILE *fp = fopen("gckeydef.default","w");
	if (fp==NULL) return;
	fprintf(fp,"DELIM :\n\nMETA @\n\n");
	fprintf(fp,"# Key Table\n\n");
	for (i=0;i<138;i++) {
		if (emptySet(keyTable[i].cmds)) continue;
		if (keyTable[i].multi)
			fprintf(fp,"MKEY %s %d %d ",pr(i),keyTable[i].pre,
					keyTable[i].post);
		else fprintf(fp,"KEY  %s ",pr(i));
		for (j=0;j<MAX_CMDS;j++)
			if (inSet(keyTable[i].cmds,j)) fprintf(fp,"%d ",j);
		if (keyTable[i].help) fprintf(fp,"\t\"%s\"",keyTable[i].help);
		fprintf(fp,"\n");
	}
	fprintf(fp,"\n\n# Command Table\n\n");
	for (i=0;i<MAX_CMDS;i++) {
		if (cmdTable[i].actions==NULL) continue;
		fprintf(fp,"%d\t: ",i);
		if (cmdTable[i].flags & C_CLR) fprintf(fp,"c");
		if (cmdTable[i].flags & C_CLRM) fprintf(fp,"C");
		if (cmdTable[i].flags & C_CATCH) fprintf(fp,"o");
		fprintf(fp,"\t: ");
		if ((cmdTable[i].restricts&F_ANY)!=F_ANY) {
			if (cmdTable[i].restricts & F_DIR) fprintf(fp,"d");
			if (cmdTable[i].restricts & F_REG) fprintf(fp,"r");
			if (cmdTable[i].restricts & F_EXEC) fprintf(fp,"x");
			if (cmdTable[i].restricts & F_LINK) fprintf(fp,"l");
			if (cmdTable[i].restricts & F_PIPE) fprintf(fp,"p");
			if (cmdTable[i].restricts & F_CDEV) fprintf(fp,"c");
			if (cmdTable[i].restricts & F_BDEV) fprintf(fp,"b");
		}
		if (cmdTable[i].regexp)
			fprintf(fp,"\t:%s:\t%s\n", cmdTable[i].regexp,
						cmdTable[i].actions);
		else
			fprintf(fp,"\t::\t%s\n", cmdTable[i].actions);
	}
	fclose(fp);
}

/***************************************************************
		DEFAULT KEY DEFINITIONS
***************************************************************/

static void
#if	__STDC__
setKeyDefaults(void) {
#else
setKeyDefaults()
{
#endif /* __STDC__ */
	keyInfo_t *f;
	f = &keyTable['q']; add2set(f->cmds,1);
	f->help = "Exit from Gram's Commander";
	f = &keyTable['Q']; add2set(f->cmds,1);
	f = &keyTable['?']; add2set(f->cmds,2);
	f->help = "Show this help";
	f = &keyTable['=']; add2set(f->cmds,55);
	f->help = "Prompt for and change the path";
	f = &keyTable['/']; add2set(f->cmds,4);
	f->help = "Start a file name incremental search";
	f = &keyTable['+']; add2set(f->cmds,5);
	f->help = "Select files using an ed(1) regular expression";
	f = &keyTable['-']; add2set(f->cmds,6);
	f->help = "Deselect files using an ed(1) regular expression";
	f = &keyTable['a']; add2set(f->cmds,7);
	f->help = "Select all files";
	f = &keyTable['n']; add2set(f->cmds,8);
	f->help = "Deselect all files";
	f = &keyTable['i']; add2set(f->cmds,9);
	f->help = "Invert selection";
	f = &keyTable['!']; add2set(f->cmds,10);
	f->help = "Read and excute a single command";
	f = &keyTable['x']; add2set(f->cmds,11);
	f->help = "Read and execute command with current file as argument";
	f = &keyTable['X']; add2set(f->cmds,12);
	f->help = "Read and do command with selected files as arguments and return";
	f = &keyTable['w']; add2set(f->cmds,13);
	f->help = "Read and do command with current file as argument and get keypress";
	f = &keyTable['W']; add2set(f->cmds,14);
	f->help = "Read and do command with selected files as arguments and get keypress";
	f = &keyTable['m']; add2set(f->cmds,15);
	f->help = "Move the current file to other directory";
	f = &keyTable['M']; add2set(f->cmds,16);
	f->multi = 1; f->pre = 54; f->post = 51;
	f->help = "Move the currently selected files to other directory";
	f = &keyTable['d']; addRange2set(f->cmds,17,18);
	f->help = "Delete the current file or directory";
	f = &keyTable['D']; addRange2set(f->cmds,19,20);
	f->multi = 1; f->post = 51; f->pre = 52;
	f->help = "Delete the selected files/directories";
	f = &keyTable['c']; add2set(f->cmds,21);
	f->help = "Copy the current file to other directory";
	f = &keyTable['C']; add2set(f->cmds,22);
	f->multi = 1; f->post = 51; f->pre = 53;
	f->help = "Copy the currently selected files to other directory";
	f = &keyTable['v']; add2set(f->cmds,23);
	f->help = "View the current file";
	f = &keyTable['V']; add2set(f->cmds,24);
	f->help = "View the currently selected files";
	f = &keyTable['e']; add2set(f->cmds,25);
	f->help = "Edit the current file";
	f = &keyTable['E']; add2set(f->cmds,26);
	f->help = "Edit the currently selected files";
	f = &keyTable['f']; add2set(f->cmds,27);
	f->help = "Run the `file' command on current file and show result";
	f = &keyTable['j']; add2set(f->cmds,28);
	f->help = "Move cursor down one line";
	f = &keyTable['k']; add2set(f->cmds,29);
	f->help = "Move cursor up one line";
	f = &keyTable['\t']; add2set(f->cmds,30);
	f->help = "Activate other window";
	f = &keyTable['U'-64]; add2set(f->cmds,31);
	f->help = "Page up";
	f = &keyTable['D'-64]; add2set(f->cmds,32);
	f->help = "Page down";
	f = &keyTable['g']; add2set(f->cmds,33);
	f->help = "Home cursor";
	f = &keyTable['G']; add2set(f->cmds,34);
	f->help = "End cursor";
	f = &keyTable[' ']; add2set(f->cmds,35);
	f->help = "Toggle current selection state and move down";
	f = &keyTable['s']; add2set(f->cmds,36);
	f->help = "Sort";
	f = &keyTable['S']; add2set(f->cmds,3);
	f->help = "Spawn a shell";
	f = &keyTable['R'-64]; add2set(f->cmds,37);
	f->help = "Repaint screen";
	f = &keyTable['r']; add2set(f->cmds,56);
	f->help = "Rename current file";
	f = &keyTable['\r']; addRange2set(f->cmds,40,49);
	f->help = "Do file dependent action on current file";
	f = &keyTable['*']; addRange2set(f->cmds,40,49);
	f->multi = 1; f->post = 51;
	f->help = "Do file dependent action on all selected files";
	f = &keyTable['F']; add2set(f->cmds,50);
	f->help = "Find files matching pattern";

	/* Set up the default commands */

	cmdTable[1 ].actions = "@i(clr) : @i(quit)";
	cmdTable[2 ].actions = "@i(clr) : @i(help) : @i(wait)";
	cmdTable[3 ].actions = "sh : @i(wait)";
	cmdTable[4 ].actions = "@i(search)";
	cmdTable[5 ].actions = "@+@ : @i(paint)";
	cmdTable[6 ].actions = "@-@ : @i(paint)";
	cmdTable[7 ].actions = "@+.@ : @i(paint)";
	cmdTable[8 ].actions = "@-.@ : @i(paint)";
	cmdTable[9 ].actions = "@i(invall) : @i(paint)";
	cmdTable[10].actions = "@aCommand? @ : @i(wait)  : @i(read): @i(paint)";
	cmdTable[11].actions = "@CCommand? @ @f : @i(wait) : @i(read): @i(paint)";
	cmdTable[12].actions = "@CCommand? @ @F : @i(wait): @i(read): @i(paint)";
	cmdTable[13].actions = "@CCommand? @ @f : @i(read): @i(paint)";
	cmdTable[14].actions = "@CCommand? @ @F : @i(read): @i(paint)";
	cmdTable[15].actions = "@qMoving@ : mv @f @O : @i(read) : @i(paint)";
	cmdTable[15].flags = C_CATCH|C_CLRM;
	cmdTable[16].actions = "mv @f @O";
	cmdTable[16].flags = C_CATCH|C_CLRM;
	cmdTable[17].actions = "@qDeleting@ : rmdir @f : @i(read) : @i(paint)";
	cmdTable[17].flags = C_CATCH;
	cmdTable[17].restricts = F_DIR;
	cmdTable[18].actions = "@qDeleting@ : @U@ : @i(read) : @i(paint)";
	cmdTable[18].restricts = F_REG|F_EXEC|F_LINK;
	cmdTable[19].actions = "rmdir @f";
	cmdTable[19].flags = C_CATCH; cmdTable[19].restricts = F_DIR;
	cmdTable[20].actions = "@U@";
	cmdTable[20].restricts = F_REG|F_EXEC|F_LINK;
	cmdTable[21].actions = "@qCopying@ : cp @f @O : @i(read) : @i(paint)";
	cmdTable[21].flags = C_CATCH|C_CLRM;
	cmdTable[22].actions = "cp @f @O";
	cmdTable[22].flags = C_CATCH|C_CLRM;
#ifdef __STDC__
	cmdTable[23].actions = "@i(clr) : @EPAGER@"PAGER"@ @f : @i(clr) : @i(paint)";
#else
	cmdTable[23].actions = "@i(clr) : @EPAGER@/bin/pg@ @f : @i(clr) : @i(paint)";
#endif
	cmdTable[23].restricts = F_REG|F_EXEC|F_LINK;
#ifdef __STDC__
	cmdTable[24].actions = "@i(clr) : @EPAGER@"PAGER"@ @F : @i(clr) : @i(paint)";
#else
	cmdTable[24].actions = "@i(clr) : @EPAGER@/bin/pg@ @F : @i(clr) : @i(paint)";
#endif
	cmdTable[24].restricts = F_REG|F_EXEC|F_LINK;
#ifdef __STDC__
	cmdTable[25].actions = "@EEDITOR@"EDITOR"@ @f : @i(clr) : @i(paint)";
#else
	cmdTable[25].actions = "@EEDITOR@/bin/vi@ @f : @i(clr) : @i(paint)";
#endif
	cmdTable[25].restricts = F_REG|F_EXEC|F_LINK;
#ifdef __STDC__
	cmdTable[26].actions = "@EEDITOR@"EDITOR"@ @F : @i(clr) : @i(paint)";
#else
	cmdTable[26].actions = "@EEDITOR@/bin/vi@ @F : @i(clr) : @i(paint)";
#endif
	cmdTable[26].restricts = F_REG|F_EXEC|F_LINK;
	cmdTable[27].actions = "file @f";
	cmdTable[27].flags = C_CATCH;
	cmdTable[28].actions = "@i(down) : @i(paint)";
	cmdTable[29].actions = "@i(up)   : @i(paint)";
	cmdTable[30].actions = "@i(swap) : @i(paint)";
	cmdTable[31].actions = "@i(pgup) : @i(paint)";
	cmdTable[32].actions = "@i(pgdn) : @i(paint)";
	cmdTable[33].actions = "@i(home) : @i(paint)";
	cmdTable[34].actions = "@i(end)  : @i(paint)";
	cmdTable[35].actions = "@i(inv)  : @i(down) : @i(paint)";
	cmdTable[36].actions = "@s@    : @i(paint)";
	cmdTable[37].actions = "@i(clr) : @i(read) : @i(paint)";
	cmdTable[40].actions = "@i(clr) : unzip -v @f : @i(wait)";
	cmdTable[40].restricts = F_REG|F_LINK;
	cmdTable[40].regexp  = "\\.zip$";
#ifdef __STDC__
	cmdTable[41].actions = "@EEDITOR@"EDITOR"@ @f : @i(clr) : @i(paint)";
#else
	cmdTable[41].actions = "@EEDITOR@/bin/vi@ @f : @i(clr) : @i(paint)";
#endif
	cmdTable[41].restricts = F_REG|F_EXEC|F_LINK;
	cmdTable[41].regexp  = "\\.[chys]$";
	cmdTable[42].actions = "make -f @f @cMake what?@ : @i(wait) : @i(read) : @i(paint)";
	cmdTable[42].restricts = F_REG|F_LINK;
	cmdTable[42].regexp  = "[Mm]akefile";
#ifdef __STDC__
	cmdTable[43].actions = "@i(clr) : @EPAGER@"PAGER"@ @f : @i(clr) : @i(paint)";
#else
	cmdTable[43].actions = "@i(clr) : @EPAGER@/bin/pg@ @f : @i(clr) : @i(paint)";
#endif
	cmdTable[43].restricts = F_REG|F_EXEC|F_LINK;
	cmdTable[43].regexp  = "[fF][aA][qQ]";
#ifdef __STDC__
	cmdTable[44].actions = "@i(clr) : @EPAGER@"PAGER"@ @f : @i(clr) : @i(paint)";
#else
	cmdTable[44].actions = "@i(clr) : @EPAGER@/bin/pg@ @f : @i(clr) : @i(paint)";
#endif
	cmdTable[44].restricts = F_REG|F_EXEC|F_LINK;
	cmdTable[44].regexp  = "[rR][eE][aA][dD].*[mM][eE]";
	cmdTable[45].actions = "@rm @f: @i(read) : @i(paint)";
	cmdTable[45].flags = C_CATCH;
	cmdTable[45].restricts = F_REG;
	cmdTable[45].regexp  = "^core$";
	cmdTable[46].actions = "strings @f | less : @i(clr) : @i(paint)";
	cmdTable[46].restricts = F_REG|F_LINK;
	cmdTable[46].regexp  = "\\.o$";
#ifdef __STDC__
	cmdTable[47].actions = "tar tvf	@f|@EPAGER@"PAGER"@ : rm @t : @i(clr) : @i(paint)";
#else
	cmdTable[47].actions = "tar tvf	@f|@EPAGER@/bin/pg@ : rm @t : @i(clr) : @i(paint)";
#endif
	cmdTable[47].restricts = F_REG|F_LINK;
	cmdTable[47].regexp  = "\\.tar$";
#ifdef __STDC__
	cmdTable[48].actions = "@EEDITOR@"EDITOR"@ @f : @i(clr) : @i(paint)";
#else
	cmdTable[48].actions = "@EEDITOR@/bin/vi@ @f : @i(clr) : @i(paint)";
#endif
	cmdTable[48].restricts = F_REG|F_EXEC|F_LINK;
	cmdTable[48].regexp  = "^[^\\.]*$";
	cmdTable[49].actions = "@i(cd) : @i(paint)";
	cmdTable[49].restricts = F_DIR;
#ifdef __STDC__
	cmdTable[50].actions = "find . -name '@cPattern? @' -print| @EPAGER@"PAGER"@ : @i(clr) :@i(paint)";
#else
	cmdTable[50].actions = "find . -name '@cPattern? @' -print| @EPAGER@/bin/pg@ : @i(clr) :@i(paint)";
#endif
	cmdTable[51].actions = "@i(read) :@i(paint)";
	cmdTable[52].actions = "@qDeleting@";
	cmdTable[53].actions = "@qCopying@";
	cmdTable[54].actions = "@qMoving@";
	cmdTable[55].actions = "@D@CChange to? @@ : @i(read) : @i(paint)";
	cmdTable[56].actions = "mv @f @CNew name for @f? @ : @i(read) : @i(paint)";
	numCmds = 57;
}


/********************************************************
	KEY DEFINITION FILE PARSER
*********************************************************/

static char
#if	__STDC__
*skipSpace(char *p) {
#else
*skipSpace(p)
	char *p;
{
#endif /* __STDC__ */
	while (*p=='\t' || *p==' ') p++;
	return p;
}

static char
#if	__STDC__
*skipPast(char *p, char c) {
#else
*skipPast(p, c)
	char *p;
	char c;
{
#endif /* __STDC__ */
	while (*p && *p!=c) p++;
	if (*p==c) p++;
	return p;
}

static char
#if	__STDC__
*getnum(char *p, int *val) {
#else
*getnum(p, val)
	char *p;
	int *val;
{
#endif /* __STDC__ */
	int v = 0;
	p = skipSpace(p);
	while (isdigit(*p)) v = v * 10 + (*p++ - '0');
	*val = v;
	return p;
}

static int
#if	__STDC__
loadKeyDefs(char *evar) {
#else
loadKeyDefs(evar)
	char *evar;
{
#endif /* __STDC__ */
	FILE *fp; char *p;
	int cmd;
	strcpy(scrap,evar);
	strcat(scrap,"/.gckeydef");
	fp = fopen(scrap,"r");
	if (fp==NULL) {
		fprintf(stderr,"Using default key assignments\n");
		return 1;
	}
	numCmds = 0;
	while (!feof(fp)) {
		fgets(scrap,1024,fp);
		cmd = strlen(scrap)-1;
		if (scrap[cmd]<32) scrap[cmd]='\0';
		if (strncmp(scrap,"DELIM ",6)==0) {
			if (p = skipSpace(scrap+6)) {
				if (*p=='\\') p++;
				delimChar = *p;
#ifdef DEBUG
				  fprintf(debug,"Set delimchar to %c\n",*p);
#endif
			}
		} else if (strncmp(scrap,"NODIRSELECT",11)==0) {
			p = skipSpace(scrap+11);
			noDirSelect = 1;
		} else if (strncmp(scrap,"META ",5)==0) {
			if (p = skipSpace(scrap+5)) {
				if (*p=='\\') p++;
				metaChar = *p; 
#ifdef DEBUG
				fprintf(debug,"Set metachar to %c\n",*p);
#endif
			}
		} else if (strncmp(scrap,"MKEY ",5)==0 || strncmp(scrap,"KEY ",4)==0) {
			int m=0, k;
			p = scrap;
			if (*p=='M') m=1;
			if (*(p = skipSpace(p+4))=='\0') continue;
			if (*p=='^') {
				if (*++p>='a') *p -= 'a'-'A';
				*p -= 64;
			} else if (*p=='F') {
				p++;
				if (*p<'0' || *p>'9')
					fprintf(stderr,"Erroneous function key definition F(%c) in keydef file!\n",*p);
				*p = *p - '0' + 128;
			} else if (*p=='\\') p++;
			if (keyTable[k=*p++].multi = m) {
				int pre, post;
				if (p = getnum(p+1,&pre))
					p = getnum(p,&post);
				keyTable[k].pre=(char)pre;
				keyTable[k].post=(char)post;
			}
#ifdef DEBUG
			if (m) fprintf(debug,"MKey %s (%d) %d %d cmds (",
				pr(k),k,keyTable[k].pre,keyTable[k].post);
			else fprintf(debug,"Key %s (%d) cmds (",pr(k),k);
#endif
			while (*p && *p!='"') {
				p = getnum(p,&cmd);
				add2set(keyTable[k].cmds,cmd);
#ifdef DEBUG
				fprintf(debug,"%d,",cmd);
#endif
			}
#ifdef DEBUG
			fprintf(debug,")\n");
#endif
			if (*(p = skipSpace(p))=='"') {
				char *ep = strrchr(scrap,'"');
				if (ep > ++p) {
					*ep = 0;
					if (keyTable[k].help = malloc(ep-p+1))
						strcpy(keyTable[k].help,p);
				}
			}
#ifdef DEBUG
			fprintf(debug,"Got key %d, help %s\n",k,p?p:"");
#endif
		} else if (isdigit(scrap[0])) {
			char restrict=0, flag=0, *re, *ep;
			p = getnum(scrap,&cmd);
			p = skipPast(p,delimChar);
			while (*p && *p!=delimChar) { /* Read flags */
				switch (*p) {
				case 'c': flag |= C_CLR;	break;
				case 'C': flag |= C_CLRM;	break;
				case 'o': flag |= C_CATCH;	break;
				}
				p++;
			}
			p = skipPast(p,delimChar);
			while (*p && *p!=delimChar) { /* Read restricts */
				switch (*p) {
				case 'd': restrict |= F_DIR;	break;
				case 'r': restrict |= F_REG;	break;
				case 'x': restrict |= F_EXEC;	break;
				case 'l': restrict |= F_LINK;	break;
				case 'p': restrict |= F_PIPE;	break;
				case 'c': restrict |= F_CDEV;	break;
				case 'b': restrict |= F_BDEV;	break;
				}
				p++;
			}
			p = skipPast(p,delimChar);
			/* Read RE */
			if ((ep = strchr(p,delimChar))==NULL) continue;
			*ep = '\0';
			re = p;
			if (*(p=ep+1)==0) continue; /* no action */
			cmdTable[cmd].restricts = restrict ? restrict : F_ANY;
			cmdTable[cmd].flags = flag;
			if (cmdTable[cmd].regexp = malloc(strlen(re)+1))
				strcpy(cmdTable[cmd].regexp,re);
			if (cmdTable[cmd].actions = malloc(strlen(p)+1))
				strcpy(cmdTable[cmd].actions,p);
#ifdef DEBUG
			fprintf(debug,"Got command %d, flag %d, restrict %d, re <%s>, act %s\n",cmd,flag,restrict,re,p);
#endif
			if (cmd>=numCmds) numCmds = cmd+1;
		} else continue;
	}
	fclose(fp);
	return 0;
}

/*****************************************************
		KEY TABLE INITIALISATION
******************************************************/

static void
#if	__STDC__
initKeyTable(char *evar) {
#else
initKeyTable(evar)
	char *evar;
{
#endif /* __STDC__ */
	int i;
	for (i=0;i<MAX_CMDS;i++) {
		cmdTable[i].flags = 0;
		cmdTable[i].restricts = F_ANY;
		cmdTable[i].regexp = NULL;
		cmdTable[i].actions = NULL;
	}
	for (i=0;i<138;i++) {
		keyTable[i].help = NULL;
		keyTable[i].multi= 0;
		keyTable[i].pre  = 0;
		keyTable[i].post = 0;
		clearSet(keyTable[i].cmds);
	}
	if (evar==NULL || loadKeyDefs(evar)) setKeyDefaults();
}

/*****************************************************************
		SAVE AND RESTORE STATE
******************************************************************/

static void
#if	__STDC__
getOldPaths(char *evar) {
#else
getOldPaths(evar)
	char *evar;
{
#endif /* __STDC__ */
	FILE *fp;
	paths[0][0] = paths[1][0] = '\0';
	if (evar==NULL) return;
	strcpy(scrap,evar); strcat(scrap,"/.gccfg");
	if (fp = fopen(scrap,"r")) {
		fscanf(fp,"%s",paths[0]);
		fscanf(fp,"%s",paths[1]);
		fclose(fp);
	}
}

static void
#if	__STDC__
savePaths(char *evar) {
#else
savePaths(evar)
	char *evar;
{
#endif /* __STDC__ */
	FILE *fp;
	if (evar==NULL) return;
	strcpy(scrap,evar); strcat(scrap,"/.gccfg");
	if (fp = fopen(scrap,"w")) {
		fprintf(fp,"%s\n",paths[0]);
		fprintf(fp,"%s\n",paths[1]);
		fclose(fp);
	}
}

/********************************************************
		FILE SELECTION
*********************************************************/

#define F_SELECTED	1

static void
#if	__STDC__
pselect(fInfo_t *p, int v) {
#else
pselect(p, v)
	fInfo_t *p;
	int v;
{
#endif /* __STDC__ */
	if (v) {
		if (noDirSelect && S_ISDIR(p->mode)) return;
		if ((p->flag & F_SELECTED)==0) {
			p->flag |= F_SELECTED;
			selCnt[l]++;
			selSize[l]+=p->size;
		}
	} else {
		if (p->flag & F_SELECTED) {
			p->flag &= ~F_SELECTED;
			selCnt[l]--;
			selSize[l]-=p->size;
		}
	}
}

static void
#if	__STDC__
nselect(int n, int v) {
#else
nselect(n, v)
	int n;
	int v;
{
#endif /* __STDC__ */
	pselect(&fInfo[l][n],v);
}

static void
#if	__STDC__
toggle(int n) {
#else
toggle(n)
	int n;
{
#endif /* __STDC__ */
	nselect(n, (fInfo[l][n].flag & F_SELECTED) ? 0 : 1);
}

/**************************************************************
		DIRECTORY SORTING
***************************************************************/

#define S_NAME		0
#define S_SIZE		1
#define S_MTIME		2
#define S_ATIME		3
#define S_OWNER		4
#define S_GROUP		5
#define S_EXTENS	6

/*
 * sortList - sort the fInfo entry indexes according to the
 *	sort keys sortStyle[l] and secSortStyle[l].
 */

static int sortingL, priStyle, secStyle, priRev, secRev;

static int
#if	__STDC__
extCmp(int *a, int *b) {
#else
extCmp(a, b)
	int *a;
	int *b;
{
#endif /* __STDC__ */
 	char *ap = strrchr(fInfo[sortingL][*a].name,'.');
 	char *bp = strrchr(fInfo[sortingL][*b].name,'.');
 	if (ap==NULL) ap=fInfo[sortingL][*a].name;
 	if (bp==NULL) bp=fInfo[sortingL][*b].name;
	return strcmp(ap,bp);
}

static int
#if	__STDC__
nameCmp(int *a, int *b) {
#else
nameCmp(a, b)
	int *a;
	int *b;
{
#endif /* __STDC__ */
	return strcmp(fInfo[sortingL][*a].name,fInfo[sortingL][*b].name);
}

static int
#if	__STDC__
sizeCmp(int *a, int *b) {
#else
sizeCmp(a, b)
	int *a;
	int *b;
{
#endif /* __STDC__ */
	return fInfo[sortingL][*b].size - fInfo[sortingL][*a].size;
}

static int
#if	__STDC__
mtimCmp(int *a, int *b) {
#else
mtimCmp(a, b)
	int *a;
	int *b;
{
#endif /* __STDC__ */
	return fInfo[sortingL][*a].modtime - fInfo[sortingL][*b].modtime;
}

static int
#if	__STDC__
atimCmp(int *a, int *b) {
#else
atimCmp(a, b)
	int *a;
	int *b;
{
#endif /* __STDC__ */
	return fInfo[sortingL][*a].acctime - fInfo[sortingL][*b].acctime;
}

static int
#if	__STDC__
ownCmp(int *a, int *b) {
#else
ownCmp(a, b)
	int *a;
	int *b;
{
#endif /* __STDC__ */
	return fInfo[sortingL][*a].uid - fInfo[sortingL][*b].uid;
}

static int
#if	__STDC__
grpCmp(int *a, int *b) {
#else
grpCmp(a, b)
	int *a;
	int *b;
{
#endif /* __STDC__ */
	return fInfo[sortingL][*a].gid - fInfo[sortingL][*b].gid;
}

typedef int (*cmpFunc)();

static cmpFunc cmpTbl[] = {
	nameCmp, sizeCmp, mtimCmp, atimCmp, ownCmp, grpCmp, extCmp
};

int
#if	__STDC__
fcmp(const void *a, const void *b) {
#else
fcmp(a, b)
	void *a;
	void *b;
{
#endif /* __STDC__ */
	int v;
	v = (*(cmpTbl[priStyle]))((int *)a,(int *)b);
	if (priRev) v = -v;
	if (v==0) {
		v = (*(cmpTbl[secStyle]))((int *)a,(int *)b);
		if (secRev) v = -v;
	}
	return v;
}

static void
#if	__STDC__
sortList(int n,char *name) {
#else
sortList(n, name)
	int n;
	char *name;
{
#endif /* __STDC__ */
	sortingL = n;
	priStyle = sortStyle[n];	priRev = reverseSort[n];
	secStyle = secSortStyle[n];	secRev = secReverseSort[n];
#ifdef __STDC__
	qsort((void *)&fIndex[n][1],numfiles[n]-1,sizeof(int),fcmp);
#else
	qsort((char *)&fIndex[n][1],numfiles[n]-1,sizeof(int),fcmp);
#endif
	if (name) {
		int i;
		for (i=0;i<numfiles[n];i++) {
			if (strcmp(name,fInfo[n][fIndex[n][i]].name)==0) {
				highlight[n] = i;
				break;
			}
		}
		normalise(n);
	} else {
		startLn[l] = highlight[l]= 0;
	}
}

/**************************************************************
	MISCELLANEOUS
***************************************************************/

static char
#if	__STDC__
*getMsg(int e) {
#else
*getMsg(e)
	int e;
{
#endif /* __STDC__ */
	char *msg;
	switch(e) {
	case EPERM:
	case EACCES:	msg = "permission denied";		break;
	case EBUSY:	msg = "busy mount point";		break;
	case EEXIST:	msg = "directory not empty";		break;
	case EINVAL:	msg = "cannot remove current directory";break;
	default:	msg = NULL;
	}
	return msg;
}

/*********************************************************
		DIRECTORY CHANGING AND READING
**********************************************************/

/*
 * readList is called after a shell command (!), startup, or
 * a change of directory. It rescans and resorts the directory.
 */

static void
#if	__STDC__
readList(int n, char *name, int newpath) {
#else
readList(n, name, newpath)
	int n;
	char *name;
	int newpath;
{
#endif /* __STDC__ */
	DIR *d = opendir(paths[n]);
	FILE *fp; char *oldname=NULL;
#ifdef DEBUG
	fprintf(debug,"In readList, ignoreDirChange=%d\n",ignoreDirChange);
#endif
	if (ignoreDirChange) return;
	if (name) strcpy(oldname=cmdline,name);
	if (d) {
		int i, j, k;
		struct dirent *de;
		struct stat st;
		/* Save old selected file names, free the rest */
		for (i=j=0;i<numfiles[n];i++) {
			if (!newpath && (fInfo[n][i].flag&F_SELECTED))
				oldSelNames[j++] = fInfo[n][i].name;
			else if (fInfo[n][i].name) 
				free(fInfo[n][i].name);
			fInfo[n][i].name = NULL;
		}
		selCnt[n] = selSize[n] = i = 0;
		while ((de=readdir(d))!=NULL) {
			if (strcmp(de->d_name,".")==0) continue;
			strcpy(scrap,paths[n]);
			strcat(scrap,"/");
			strcat(scrap,de->d_name);
			if (stat(scrap,&st)==0) {
				register fInfo_t *f = &fInfo[n][i];
				f->size = st.st_size;
				f->mode = st.st_mode;
				f->uid = st.st_uid;
				f->gid = st.st_gid; 
				f->modtime = st.st_mtime;
				f->acctime = st.st_atime;
				{ register mode_t m = f->mode;
				if (S_ISDIR(m)) f->typID = '/';
				else if (S_ISCHR(m)) f->typID = 'c';
				else if (S_ISBLK(m)) f->typID = 'b';
				else if (S_ISFIFO(m)) f->typID = 'p';
				else if (S_ISLNK(m)) f->typID = 'l';
				else if (m & (S_IXUSR|S_IXGRP|S_IXOTH))
					f->typID = '*';
				else f->typID = ' ';
				}
				f->flag = 0;
				for (k=0;k<j;k++) {
					if (strcmp(oldSelNames[k],de->d_name)==0) {
						f->flag = F_SELECTED;
						selCnt[n]++;
						selSize[n] += f->size;
						break;
					}
				}
				if (f->name = malloc(strlen(de->d_name)+1))
					strcpy(f->name,de->d_name);
				fIndex[n][i] = i;
				i++;
				if (i>=MAXARGS) break;
			} else {
				/* ERROR */
			}
		}
		numfiles[n] = i;
		if (newpath) highlight[n]=0;
		else if (highlight[n]>=numfiles[n])
			highlight[n] = numfiles[n]-1;
		closedir(d);
		if (!newpath) 
			while (--j >= 0)
				free(oldSelNames[j]);
	} else {
		/* ERROR !!!! */
	}
	sortList(n,oldname);
}

static void
#if	__STDC__
reScan(char *name) {
#else
reScan(name)
	char *name;
{
#endif /* __STDC__ */
	getcwd(paths[l],512+256);
	readList(l,name,1);
	showPath(l);
}

static void
#if	__STDC__
doCD(char *d, char *name) {
#else
doCD(d, name)
	char *d;
	char *name;
{
#endif /* __STDC__ */
#ifdef DEBUG
	fprintf(debug,"In doCD(%s), ignoreDirChange=%d\n",d,ignoreDirChange);
#endif
	if (chdir(d)==0) { reScan(name); clearCmdWin(); }
	else {
		char *msg = getMsg(errno);
		mvwprintw(cmdWin,0,0,"Failed to change to directory %s %c %s",cmdline+4,msg?'-':' ',msg?msg:"");
		UPDATE(cmdWin);
		REFRESH();
		reScan(fname);
	}
}

/***********************************************************
		SCREEN HANDLING
************************************************************/

static void
#if	__STDC__
showCursor(int n, int sel) {
#else
showCursor(n, sel)
	int n;
	int sel;
{
#endif /* __STDC__ */
	mvwprintw(panels[n],highlight[n]-startLn[n],1,sel?"=>":"->");
	UPDATE(panels[n]);
}

static void
#if	__STDC__
hideCursor(int n) {
#else
hideCursor(n)
	int n;
{
#endif /* __STDC__ */
	mvwprintw(panels[n],highlight[n]-startLn[n],1,"  ");
	UPDATE(panels[n]);
}

static void
#if	__STDC__
clearCmdWin(void) {
#else
clearCmdWin()
{
#endif /* __STDC__ */
	werase(cmdWin);
	if (searching) mvwprintw(cmdWin,0,0,"/%s",srchTxt);
	UPDATE(cmdWin);
	REFRESH();
}

static void
#if	__STDC__
showCmdText(char *txt) {
#else
showCmdText(txt)
	char *txt;
{
#endif /* __STDC__ */
	werase(cmdWin);
	mvwprintw(cmdWin,0,0,txt);
	UPDATE(cmdWin);
	REFRESH();
}

static int
#if	__STDC__
getInput(char *prompt, char *buf, int len) {
#else
getInput(prompt, buf, len)
	char *prompt;
	char *buf;
	int len;
{
#endif /* __STDC__ */
	showCmdText(prompt);
	echo();
	WGETNSTR(cmdWin,buf,len);
	noecho();
	clearCmdWin();
	return strlen(buf);
}

static void
#if	__STDC__
showTitle(void) {
#else
showTitle()
{
#endif /* __STDC__ */
	WINDOW *w = newwin(13,27,(HEIGHT-14)/2,(COLS-24)/2);
	box(w,0,0);
	mvwprintw(w,2,6,"Gram's Commander");
	mvwprintw(w,4,12,"by");
	mvwprintw(w,6,6,"Graham Wheeler");
	mvwprintw(w,8,3,"gram@aim1.aztec.co.za");
	mvwprintw(w,10,11,"v2.2");
	UPDATE(w); REFRESH();
	sleep(2);
	wclear(w); UPDATE(w); delwin(w); REFRESH();
}

static void
#if	__STDC__
setupCurses(int first) {
#else
setupCurses(first)
int first;
{
#endif /* __STDC__ */
	if (inCurses) return;
	initscr(); nonl(); raw(); noecho();
	if (first) { i_clr(); showTitle(); }
	panelTop[0] = newwin(NUM_FILES+2,COLS/2,0,0);
	panels[0] = subwin(panelTop[0],NUM_FILES,COLS/2-2,1,1);
	panelTop[1] = newwin(NUM_FILES+2,COLS/2,0,COLS/2);
	panels[1] = subwin(panelTop[1],NUM_FILES,COLS/2-2,1,COLS/2+1);
	statusTop[0] = newwin(NUM_INFO_LNS+2,COLS/2,LINES-NUM_INFO_LNS-3,0);
	statusWin[0] = subwin(statusTop[0],NUM_INFO_LNS,COLS/2-2,LINES-NUM_INFO_LNS-2,1);
	statusTop[1] = newwin(NUM_INFO_LNS+2,COLS/2,LINES-NUM_INFO_LNS-3,COLS/2);
	statusWin[1] = subwin(statusTop[1],NUM_INFO_LNS,COLS/2-2,LINES-NUM_INFO_LNS-2,COLS/2+1);
	cmdWin = newwin(1,COLS,LINES-1,0);
	KEYPAD(panels[0],1);
	KEYPAD(panels[1],1);
	KEYPAD(cmdWin,1);
	inCurses = 1;
}

/*
 * normalise() is called after the cursor is moved, to ensure
 * the cursor is still displayed in the window, etc.
 */

static void
#if	__STDC__
normalise(int n) {
#else
normalise(n)
	int n;
{
#endif /* __STDC__ */
	if ((highlight[n]-startLn[n])>=NUM_FILES) {
		startLn[n] = highlight[n]-NUM_FILES+1;
	} else if (highlight[n]<startLn[n]) startLn[n] = highlight[n];
	if (startLn[n]>=numfiles[n]) startLn[n] = numfiles[n]-1;
	if (highlight[n]>=numfiles[n]) highlight[n] = numfiles[n]-1;
}

static void
#if	__STDC__
showPath(int p) {
#else
showPath(p)
	int p;
{
#endif /* __STDC__ */
	int L = strlen(paths[p]);
	if (L>30) L-=30; else L=0;
	box(panelTop[p],0,0);
	if (p==l) wstandout(panelTop[p]);
	mvwprintw(panelTop[p],0,5,"<%s>",paths[p]+L);
	wstandend(panelTop[p]);
	UPDATE(panelTop[p]);
}

static void
#if	__STDC__
showStatus(int n) {
#else
showStatus(n)
	int n;
{
#endif /* __STDC__ */
	char perms[14];
	struct group *grp;
	struct passwd *usr;
	struct tm  *tp;
	int i=0, F = fIndex[n][highlight[n]];
	box(statusTop[n],0,0);
	UPDATE(statusTop[n]);
	mvwprintw(statusWin[n],0,1,"%3d files selected (%-9ld bytes)",
		selCnt[n],selSize[n]);
	perms[i++] = (fInfo[n][F].mode&S_ISUID) ? 's' : '-';
	perms[i++] = (fInfo[n][F].mode&S_ISGID) ? 'S' : '-';
	perms[i++] = (fInfo[n][F].mode&S_ISVTX) ? 't' : '-';
	perms[i++] = (fInfo[n][F].mode&S_IRUSR) ? 'r' : '-';
	perms[i++] = (fInfo[n][F].mode&S_IWUSR) ? 'w' : '-';
	perms[i++] = (fInfo[n][F].mode&S_IXUSR) ? 'x' : '-';
	perms[i++] = (fInfo[n][F].mode&S_IRGRP) ? 'r' : '-';
	perms[i++] = (fInfo[n][F].mode&S_IWGRP) ? 'w' : '-';
	perms[i++] = (fInfo[n][F].mode&S_IXGRP) ? 'x' : '-';
	perms[i++] = (fInfo[n][F].mode&S_IROTH) ? 'r' : '-';
	perms[i++] = (fInfo[n][F].mode&S_IWOTH) ? 'w' : '-';
	perms[i++] = (fInfo[n][F].mode&S_IXOTH) ? 'x' : '-';
	perms[i++]=0;
	grp = getgrgid(fInfo[n][F].gid);
	usr = getpwuid(fInfo[n][F].uid);
	mvwprintw(statusWin[n],1,1," %9s  %10s/%-10s",
		perms, usr->pw_name, grp->gr_name);
	mvwprintw(statusWin[n],2,1," Modified %s",
			asctime(localtime(&fInfo[n][F].modtime)));
	mvwprintw(statusWin[n],3,1," Accessed %s",
			asctime(localtime(&fInfo[n][F].acctime)));
	UPDATE(statusWin[n]);
}

/*
 * showInfo - display the i'th entry for the n panel on the
 *	r'th row of the panel window
 */

static void
#if	__STDC__
showInfo(int n, int r, int i) {
#else
showInfo(n, r, i)
	int n;
	int r;
	int i;
{
#endif /* __STDC__ */
	char buff[24]; int F = fIndex[n][i];
	mvwprintw(panels[n],r,4,"                                  ");
	if (i>=numfiles[n]) return;
	strncpy(buff,fInfo[n][F].name,22); buff[22]=0;
	if (fInfo[n][F].flag&F_SELECTED) wstandout(panels[n]);
	mvwprintw(panels[n],r,4,"%c %s",fInfo[n][F].typID,buff);
	wstandend(panels[n]);
	mvwprintw(panels[n],r,30,"%8ld",fInfo[n][F].size);
}

static void
#if	__STDC__
showList(int n) {
#else
showList(n)
	int n;
{
#endif /* __STDC__ */
	int i;
	for (i=0;i<NUM_FILES;i++) showInfo(n,i,startLn[n]+i);
	UPDATE(panels[n]);
}

/************************************************************
		INTERNAL PRIMITIVE COMMANDS
*************************************************************/

static void
#if	__STDC__
i_read(void) {
#else
i_read()
{
#endif /* __STDC__ */
	int i;
	for (i=0;i<2;i++)
		readList(i,fInfo[i][fIndex[i][highlight[i]]].name,0);
}

static void
#if	__STDC__
i_clr(void) {
#else
i_clr()
{
#endif /* __STDC__ */
	wclear(stdscr);
	UPDATE(stdscr);
	REFRESH();
}

static void
#if	__STDC__
i_wait(void) {
#else
i_wait()
{
#endif /* __STDC__ */
	fprintf(stderr,"Press ENTER to return to Gram's Commander...");
	fflush(stderr);
	wgetch(cmdWin);
	i_clr(); i_paint();
}

static void
#if	__STDC__
i_paint(void) {
#else
i_paint()
{
#endif /* __STDC__ */
	int i;
#ifdef DEBUG
	fprintf(debug,"In i_paint, ignoreDirChange=%d\n",ignoreDirChange);
#endif
	for (i=0;i<2;i++) {
		normalise(i);
		showPath(i);
		showStatus(i);
		showList(i);
		/*showCursor(i,i==l);*/
	}
	/*UPDATE(curscr);*/
	REFRESH();
}

static void
#if	__STDC__
i_quit(void) {
#else
i_quit()
{
#endif /* __STDC__ */
	i_clr();
	endwin();
	savePaths(getenv("GCDIR"));
#ifdef DEBUG
	fclose(debug);
#endif
	exit(0);
}

static void
#if	__STDC__
i_help(void) {
#else
i_help()
{
#endif /* __STDC__ */
	int i, r;
	char *p;
	for (i=r=0;i<138;i++) {
		if (emptySet(keyTable[i].cmds)) continue;
		if (keyTable[i].help==NULL) continue;
		if (r==0) 
			fprintf(stderr,"\n\rKey\tAction\r\n===\t============================================================\n\r");
			
		r++;
		if (i==' ') fprintf(stderr,"Space\t");
		else if (i==9) fprintf(stderr,"Tab\t");
		else if (i==13) fprintf(stderr,"Enter\t");
		else if (i<32) fprintf(stderr,"^%c\t",i+64);
		else if (i<128) {
			if (i=='F' || i=='\\' || i=='^')
				fprintf(stderr,"\\%c\t",i);
			else 
				fprintf(stderr,"%c\t",i);
		} else fprintf(stderr,"F%-2d\t",i-128);
		if ((p=keyTable[i].help)==NULL) p="";
		fprintf(stderr,"%-s\n\r",p);
		if (r==NUM_FILES) {
			fprintf(stderr,"\nPress a key...");
			fflush(stderr);
			r = 0;
			getch();
		}
	}
}

static void
#if	__STDC__
i_invall(void) {
#else
i_invall()
{
#endif /* __STDC__ */
	int i;
	for (i=0;i<numfiles[l];i++) toggle(i);
}

static void
#if	__STDC__
cd2old(void) {
#else
cd2old()
{
#endif /* __STDC__ */
	chdir(paths[l]);
	/*readList(l,fInfo[l][fIndex[l][highlight[l]]].name,0);*/
	showPath(l);
	clearCmdWin();
}

static void
#if	__STDC__
i_inv(void) {
#else
i_inv()
{
#endif /* __STDC__ */
	toggle(f);
}

static void
#if	__STDC__
i_search(void) {
#else
i_search()
{
#endif /* __STDC__ */
	searching = 1;
}

static void
#if	__STDC__
i_swap(void) {
#else
i_swap()
{
#endif /* __STDC__ */
	if (!ignoreDirChange) {
		l=1-l; cd2old();
	}
}

static void
#if	__STDC__
i_down(void) {
#else
i_down()
{
#endif /* __STDC__ */
	if (highlight[l]<(numfiles[l]-1)) {
		highlight[l]++;
		normalise(l);
	}
}

static void
#if	__STDC__
i_up(void) {
#else
i_up()
{
#endif /* __STDC__ */
	if (highlight[l]>0) {
		highlight[l]--;
		normalise(l);
	}
}

static void
#if	__STDC__
i_pgup(void) {
#else
i_pgup()
{
#endif /* __STDC__ */
	if (highlight[l]!=startLn[l]) highlight[l] = startLn[l];
	else {
		int n = NUM_FILES;
		if (n>highlight[l]) n = highlight[l];
		startLn[l] = (highlight[l] -= n);
	}
}

static void
#if	__STDC__
i_pgdn(void) {
#else
i_pgdn()
{
#endif /* __STDC__ */
	if (highlight[l]!=(startLn[l]+NUM_FILES-1))
		highlight[l] = startLn[l]+NUM_FILES-1;
	else 	{
		if ((numfiles[l]-startLn[l])>=NUM_FILES) {
			startLn[l] += NUM_FILES;
			highlight[l] = startLn[l]+NUM_FILES-1;
		}
	}
	normalise(l);
}

static void
#if	__STDC__
i_home(void) {
#else
i_home()
{
#endif /* __STDC__ */
	highlight[l] = 0;
}

static void
#if	__STDC__
i_end(void) {
#else
i_end()
{
#endif /* __STDC__ */
	highlight[l] = numfiles[l] - 1;
}

static void
#if	__STDC__
i_beep(void) {
#else
i_beep()
{
#endif /* __STDC__ */
	BEEP;
}

static void
#if	__STDC__
i_cd(void) {
#else
i_cd()
{
#endif /* __STDC__ */
	char *oldname=NULL;
	if (ignoreDirChange) return;
	if (strcmp(fname,"..")==0) {
		char *p = strrchr(paths[l],'/');
		strcpy(scrap,p+1);
		if (p==paths[l]) { /* root dir */
			if (p[1]) p[1] = '\0';
			else i_beep();
		} else *p = '\0';
		oldname = scrap;
	}  else if (strcmp(fname,".")!=0) {
		if (paths[l][1]) strcat(paths[l],"/");
		strcat(paths[l],fname);
	}
	doCD(paths[l],oldname);
}

/****************************************************
		REGULAR EXPRESSION MATCHING
*****************************************************/

#if defined (USE_SYSV_REGEXP) || defined(USE_BSD_REGEXP)

#ifdef USE_SYSV_REGEXP
static char REbuf[1024], *REmsg, *REsrc;

#define INIT		register char *sp=REsrc;
#define GETC()		(*sp++)
#define UNGETC(c)	(sp--)
#define RETURN(c)	return;
#define ERROR		REerr
#define PEEKC()		(*sp)

static void
REerr(int c) {
	switch(c) {
	case 11: REmsg = "range endpoint too large"; break;
	case 16: REmsg = "bad number"; break;
	case 25: REmsg = "\\ digit out of range"; break;
	case 36: REmsg = "illegal or missing delimeter"; break;
	case 41: REmsg = "no remembered search string"; break;
	case 42: REmsg = "\\( \\) imbalance"; break;
	case 43: REmsg = "too many \\('s"; break;
	case 44: REmsg = "more than two numbers in \\{ \\}"; break;
	case 45: REmsg = "} expected after \\"; break;
	case 46: REmsg = "first number exceeds second in \\{ \\}"; break;
	case 49: REmsg = "[ ] imbalance"; break;
	case 50: REmsg = "regular expression overflow"; break;
	default: REmsg = "unknown error"; break;
	}
}

#include <regexp.h>

#endif

static char
#ifdef __STDC__
*compileRE(char *RE) {
#else
*compileRE(RE)
	char *RE;
{
#endif
#ifdef USE_BSD_REGEXP
	return re_comp(RE);
#else
	REmsg = NULL;
	REsrc = RE;
	compile(NULL,REbuf,REbuf+1024,'\0');
	return REmsg;
#endif
}

static int
#ifdef __STDC__
matchRE(char *string) {
#else
matchRE(string)
	char *string;
{
#endif
#ifdef USE_BSD_REGEXP
	return (re_exec(string)==1);
#else
	return (step(string,REbuf));
#endif
}
#else
static char
#if	__STDC__
*compileRE(char *RE) {
#else
*compileRE(RE)
	char *RE;
{
#endif /* __STDC__ */
	return NULL;
}
static int
#if	__STDC__
matchRE(char *string) {
#else
matchRE(string)
	char *string;
{
#endif /* __STDC__ */
	return 0;
}
#endif

/***************************************************************
		KEY ACTION PROCESSING
****************************************************************/

#define badCommand(n)	{Enum = n; goto PARSEFAIL; }

static char actionBuf[1024], inputBuf[1024], tmpBuf[1024];

/*
 * preProcess handles any embedded placeholders
 */

static int
#if	__STDC__
preProcess(int pass, char *p, fInfo_t *ip) {
#else
preProcess(pass, p, ip)
	int pass;
	char *p;
	fInfo_t *ip;
{
#endif
	char ebuf[25], *p2 = p-1, *insrt, *head, *tail,
		*evar, *defalt, *prompt;
	int Enum, i;
	while (p2 = strchr(p2+1,metaChar)) {
		head = p2++;
		tail = p2+1;
#ifdef DEBUG
		fprintf(debug,"preProcessing %s from %s\n",p,head);
#endif
		switch (*p2) {
		case 'E': 
			if (pass==0) goto SKIP;
			p2 = strchr((evar = p2+1), metaChar);
			if (p2 && *p2) *p2 = '\0';
			else badCommand(7);
			p2 = strchr((defalt = p2+1), metaChar);
			if (p2 && *p2) *p2 = '\0';
			else badCommand(7);
			tail = p2+1;
			if ((insrt = getenv(evar)) == NULL)
				insrt = defalt;
			break;
		case 'c': 
		case 'C': 
			if (pass==0) goto SKIP;
			Enum = (int)*p2;
			if (*++p2=='\0') badCommand(8);
			if (*p2!=metaChar) {
				prompt = p2;
				if (!(p2=strchr(p2,metaChar)))
					badCommand(9);
				*p2 = '\0';
			} else prompt = "? ";
			if (getInput(prompt,inputBuf,80)==0 && Enum=='c')
				 return 1;;
			tail = p2+1;
			insrt = inputBuf;
			break;
		case 'F':
			inputBuf[0] = '\0';
			for (i=0;i<numfiles[l];i++)
				if (fInfo[l][i].flag & F_SELECTED) {
					/* No bounds checking! */
					strcat(inputBuf,fInfo[l][i].name);
					strcat(inputBuf," ");
				}
			insrt = inputBuf;
			break;
		case 'l': insrt = getlogin();	break;
		case 't': insrt = tempName;	break;
		case 'f': insrt = ip->name;	break;
		case 'p': insrt = paths[l];	break;
		case 'O': insrt = paths[1-l];	break;
		case 'o': insrt = fInfo[1-l][fIndex[1-l][highlight[1-l]]].name;
						break;
		default:
		SKIP:
			p2--; continue;
		}
		/* Now we have to piece together the head, insrt and tail */
		*head = '\0';
		strcpy(tmpBuf, p);
		p2 = p + strlen(tmpBuf) - 1;
#ifdef DEBUG
		fprintf(debug,"Doing substitution: head <%s>\n\tinsrt <%s>\n\ttail <%s>\n",p,insrt,tail);
#endif
		if (insrt) strcat(tmpBuf,insrt);
		if (tail) strcat(tmpBuf,tail);
		strcpy(p,tmpBuf);
	}
	return 0;
PARSEFAIL:
	sprintf(ebuf, "Error in command! (%d)",Enum);
	showCmdText(ebuf);
	return -1;
}

static int
#if	__STDC__
doAction(int flags, char *buff, fInfo_t *ip) {
#else
doAction(flags, buff, ip)
	int flags;
	char *buff;
	fInfo_t *ip;
{
#endif /* __STDC__ */
	char *p, *op, *tmp, ebuf[25]; int i,Enum;
	strcpy(p=op=actionBuf,buff);
	if (preProcess(0, p, ip)) return 1;
	if (preProcess(1, p, ip)) return 1;
	cmdline[0]='\0';
	strcpy(ebuf,"Error in command! ");
#ifdef DEBUG
	fprintf(debug,"In doAction(%s)\n",p);
#endif
	while (p = strchr(p,metaChar)) {
		*p='\0';
		strcat(cmdline,op);
		if (p[1]=='i') {
			if (p[2]=='(') { /* internal cmd */
				i = 0;
				op = p+3;
				if (!op) badCommand(1);
				p = strchr(op,')');
				if (!p) badCommand(2);
				while (--p, (*p==' ' || *p=='\t') && p>op);
				p[1] = '\0';
#ifdef DEBUG
				fprintf(debug,"Looking up internal cmd %s\n",op);
#endif
				while (iCmdNames[i]) {
					if (strcmp(op,iCmdNames[i])==0) {
#ifdef DEBUG
						fprintf(debug,"Doing internal command %s\n",iCmdNames[i]);
#endif
						(*(iCmdTbl[i]))();
						return 0;
					}
					i++;
				}
			}
			badCommand(3);
		} else if (p[1]=='+' || p[1]=='-') {
			int v = (p[1]=='+');
			char re[80], *dbg;
			if (p[2]==metaChar) {
				if (getInput("Pattern? ",re,80)==0) return 1;
			} else {
				if (p=strchr(op=p+2,metaChar))  *p = '\0';
				else badCommand(4);
				strcpy(re,op);
			}
			if ((dbg=compileRE(re))==NULL) {
				for (i=0;i<numfiles[l];i++) {
					if (matchRE(fInfo[l][i].name)==1)
						nselect(i,v);
				}
				return 0;
			} else {
				showCmdText(dbg);
				return 1;
			}
		} else if (p[1]=='D') { /* Change directory */
			op = strchr(++p,metaChar);
			if (op) {
				*op++ = '\0'; ++p;
				if (strcmp(p,"..")==0) {
					if (strcmp(paths[l],"/")) {
						char oldname[256];
						p = strrchr(paths[l],'/');
						strcpy(oldname,p+1);
						if (p==paths[l]) p++;
						*p = '\0';
						doCD(paths[l],oldname);
					}
				} else if (*p=='/') {
					strcpy(paths[l],p);
					doCD(paths[l],NULL);
				} else if (strcmp(p,".")!=0) {
					if (paths[l][1]) strcat(paths[l],"/");
					strcat(paths[l],p);
					doCD(paths[l],NULL);
				}
				return 0;
			} else badCommand(5);
		} else if (p[1]=='U') { /* Unlink file */
			op = strchr(++p,metaChar);
			if (op) {
				int rtn; char *n;
				if (op == ++p) rtn = unlink(n=ip->name);
				else {
					*op = '\0';
					rtn = unlink(n=p);
				}
				if (rtn) {
					char msg[80];
					sprintf(msg,"Unlink of %.60s failed",n);
					showCmdText(msg);
				} else readList(l, fname, 0);
				return 0;
			} else badCommand(6);
		} else if (p[1]=='q') { /* confirm */
			op = strchr(++p,metaChar);
			if (op) {
				int ch;
				*op++ = '\0';
				clearCmdWin();
#ifdef DEBUG
				fprintf(debug,"flags %d, *p=%s\n",flags,pr(*p));
#endif
				if ((flags & C_MULTI) && !inMultiLoop) {
					if (selCnt[l]) 
						mvwprintw(cmdWin,0,0,"%s %d bytes in %d files - confirm y/n? ",
							p+1, selSize[l],selCnt[l]);
					else return 0;
				} else mvwprintw(cmdWin,0,0,"%s %d bytes in file %s - confirm y/n? ",
						p+1,ip->size,ip->name);
				UPDATE(cmdWin); REFRESH();
				ch=wgetch(cmdWin);
				clearCmdWin();
				return (ch=='y'|| ch=='Y')?0:1;
			}
			badCommand(7);
		} else if (p[1]=='s') {
			int key;
			secReverseSort[l] = reverseSort[l];
			secSortStyle[l] = sortStyle[l];
			if (p[2]==metaChar) {
				showCmdText("Sort key?");
				ch = wgetch(panels[l]);
				clearCmdWin();
			} else ch=p[2];
			if (isascii(ch) && isupper(ch)) {
				reverseSort[l] = 1;
				ch = tolower(ch);
			} else reverseSort[l] = 0;
			switch(ch) {
			case 'a': sortStyle[l] = S_ATIME; break;
			case 'm': sortStyle[l] = S_MTIME; break;
			case 'n': sortStyle[l] = S_NAME; break;
			case 's': sortStyle[l] = S_SIZE; break;
			case 'g': sortStyle[l] = S_GROUP; break;
			case 'u': sortStyle[l] = S_OWNER; break;
			case 'x': sortStyle[l] = S_EXTENS; break;
			}
			sortList(l,fname);
			/*showList(l);*/
			return 0;
		} else if (p[1]=='a') {
			p += 2;
			if (*p=='\0') badCommand(8);
			if (*p!=metaChar) {
				if (!(op=strchr(p,metaChar)))
					badCommand(9);
				*op = '\0';
			} else {
				op = p;
				p = "? ";
			}
			if (getInput(p,inputBuf,80)==0) return 1;
			else return doAction(flags,inputBuf,ip);
		} else badCommand(99);
	}
	strcat(cmdline,op);
	/* Now we execute the command, taking the flags into account */
	if (flags&C_CATCH) {
		tmp = tmpnam(NULL);
		strcat(cmdline," >");
		strcat(cmdline,tmp);
		strcat(cmdline," 2>&1");
	} else endwin();
#ifdef DEBUG
	fprintf(debug,"SHELLING %s\n",cmdline);
#endif
	if (system(cmdline)==0) 
		if ( (flags & (C_CLR|C_MULTI)) == C_CLR ||
			(flags & (C_CLRM|C_MULTI)) == (C_CLRM|C_MULTI))
				pselect(ip,0);
	if (flags&C_CATCH) {
		FILE *fp = fopen(tmp,"r");
		if (fp) {
			scrap[0]='\0';
			fgets(scrap,78,fp); scrap[78]='\0';
			fclose(fp);
		} else strcpy(scrap,"No output!");
		unlink(tmp);
		showCmdText(scrap);
	} else setupCurses(0);
	return 0;
PARSEFAIL:
	sprintf(ebuf, "Error in command! (%d)",Enum);
#ifdef DEBUG
	fprintf(debug, "Error in command! (%d)\nbuff=%s\np = %s\nop=%s\n",
		Enum,buff,p,op);
#endif
	showCmdText(ebuf);
	return -1;
}

static int
#if	__STDC__
doCommand(int cmd, int multi, fInfo_t *ip) {
#else
doCommand(cmd, multi, ip)
	int cmd;
	int multi;
	fInfo_t *ip;
{
#endif /* __STDC__ */
	char *ep, *p=cmdTable[cmd].actions-1, buff[256];
	while (p) {
		int L;
		ep = strchr(++p,delimChar);
		L = ep ? (ep-p) : strlen(p);
		if (L>255) L=255;
		strncpy(buff,p,L); buff[L] = '\0';
		if (doAction(cmdTable[cmd].flags | (multi?C_MULTI:0),
			buff,ip)) {
				/* Something failed */
				i_paint();
				return 1;
		}
		p = ep;
	}
	return 0;
}

static int
#if	__STDC__
applyKey(int ch, fInfo_t *ip, int multi) {
#else
applyKey(ch, ip, multi)
	int ch;
	fInfo_t *ip;
	int multi;
{
#endif /* __STDC__ */
	int cmd, t;
	char *re;
#ifdef DEBUG
	fprintf(debug,"Attempting to apply %skey %s (%d)\n",multi?"multi":"",pr(ch),ch);
#endif
	for (cmd=0;cmd<numCmds;cmd++) {
		if (!inSet(keyTable[ch].cmds,cmd)) continue;
		if (cmdTable[cmd].actions==NULL) continue;
#ifdef DEBUG
		fprintf(debug,"Trying command %d\n",cmd);
#endif
		if ((t = cmdTable[cmd].restricts) != F_ANY) {
#ifdef DEBUG
			fprintf(debug,"Checking restrictions %d\n",
				cmdTable[cmd].restricts);
#endif
			if (S_ISREG(ip->mode)) {
				if ((t&F_REG)==0)
					if ((ip->mode&(S_IXUSR|S_IXGRP|S_IXOTH))==0)
						continue;
				if ((t&F_EXEC)==0)
					if (ip->mode&(S_IXUSR|S_IXGRP|S_IXOTH))
						continue;
			}
			if ((t&F_DIR)==0)
				if (S_ISDIR(ip->mode)) continue;
			if ((t&F_PIPE)==0)
				if (S_ISFIFO(ip->mode)) continue;
			if ((t&F_CDEV)==0)
				if (S_ISCHR(ip->mode)) continue;
			if ((t&F_BDEV)==0)
				if (S_ISBLK(ip->mode)) continue;
			if ((t&F_LINK)==0)
				if (S_ISLNK(ip->mode)) continue;
		}
#ifdef DEBUG
		fprintf(debug,"Passed restrictions; checking RE\n");
#endif
		if ((re = cmdTable[cmd].regexp) && *re) {
			if (compileRE(re)==NULL)
				if (!matchRE(ip->name)) continue;

		}
		/* MATCHED! Now we apply the actions */
#ifdef DEBUG
		fprintf(debug,"PASSED key %d - doing actions in cmd %d\n",ch,cmd);
		fprintf(debug,"key actions: %s\n",cmdTable[cmd].actions);
#endif
		return doCommand(cmd, multi, ip);
	}
	return 0;
}

/***********************************************************
		INCREMENTAL FILE NAME SEARCH
***********************************************************/

static int
#if	__STDC__
srchMatch(int n, int ch) {
#else
srchMatch(n, ch)
	int n;
	int ch;
{
#endif /* __STDC__ */
	int L = strlen(srchTxt);
	int H = highlight[n];
	srchTxt[L]=ch; srchTxt[L+1]='\0';
	do {
		if (strncmp(fInfo[n][fIndex[n][H]].name,srchTxt,L+1)==0) {
			highlight[n] = H;
			normalise(n);
			i_paint();
			return 1;
		}
		H++;
		if (H==numfiles[n]) H=0;
	} while (H != highlight[n]);
	srchTxt[L] = '\0';
	return 0;
}

/**************************************************************
		MAIN PROGRAM
**************************************************************/

static void
#if	__STDC__
loop(void) {
#else
loop()
{
#endif /* __STDC__ */
	srchTxt[0]='\0';
	for (;;) {
		if (searching) clearCmdWin();
		showCursor(1-l,0);
		showCursor(l,1);
		REFRESH();
		ch = wgetch(panels[l]);
		/* hideCursor(1-l); */ hideCursor(l);
		f = fIndex[l][highlight[l]];
		fname = fInfo[l][f].name;
		if (searching) {
			if (isascii(ch) && ch>32) {
				if (!srchMatch(l, ch)) i_beep();
				continue;
			} else {
				searching = 0;
				srchTxt[0]='\0';
				clearCmdWin();
			}
		}
		switch (ch) {
			case KEY_UP:	i_up(); i_paint(); continue;
			case KEY_DOWN:	i_down(); i_paint(); continue;
			case KEY_LEFT:	l=0; cd2old(); i_paint(); continue;
			case KEY_RIGHT:	l=1; cd2old(); i_paint(); continue;
			case KEY_PPAGE: i_pgup(); i_paint(); continue;
			case KEY_NPAGE: i_pgdn(); i_paint(); continue;
			case KEY_HOME:	i_home(); i_paint(); continue;
			case KEY_END:	i_end(); i_paint(); continue;
			case KEY_HELP:  i_help(); i_clr(); i_paint(); continue;
			default: if (!isascii(ch)) {
					if (ch>=KEY_F0 && ch<(KEY_F0+10))
						ch = 128 + ch - KEY_F0;
					else	{
						i_beep();
						continue;
					}
				}
		}
		if (emptySet(keyTable[ch].cmds)) {
			i_beep();
#ifdef DEBUG
			fprintf(debug,"Got keypress %d but entry has no commands\n",ch);
#endif
			continue;
		}
		if (keyTable[ch].multi) {
			int i=0, k=ch;
			ignoreDirChange = 1;
			if (keyTable[k].pre)
				if (doCommand(keyTable[k].pre,1,NULL))
					goto skipLoop;
			inMultiLoop = 1;		
			for (i=0; i<numfiles[l]; i++) {
#ifdef DEBUG
					fprintf(debug,"Checking file %d %s\n",i,fInfo[l][i].name);
#endif
				if (fInfo[l][i].flag & F_SELECTED) {
#ifdef DEBUG
					fprintf(debug,"Calling applykey with file %s\n",fInfo[l][i].name);
#endif
					applyKey(k, &fInfo[l][i], 1);
				}
			}
			inMultiLoop = ignoreDirChange = 0;
			if (keyTable[k].post)
				doCommand(keyTable[k].post,1,NULL);
skipLoop:;
		} else {
			inMultiLoop = ignoreDirChange = 0;
			applyKey(ch,&fInfo[l][f], 0);
		}
	}
}

static void
#if	__STDC__
useage(char *name) {
#else
useage(name)
	char *name;
{
#endif /* __STDC__ */
	fprintf(stderr,"Gram's Commander v2.0\n");
	fprintf(stderr,"Useage: %s [ <path> [ <path> ] ]\n",name);
	fprintf(stderr,"or:     %s -D  to dump default assignments to `gckeydef.default'\n",name);
}

#if	__STDC__
main(int argc, char *argv[]) {
#else
main(argc, argv)
	int argc;
	char **argv;
{
#endif /* __STDC__ */
	char *evar = getenv("GCDIR");
	getOldPaths(evar);
	if (argc>1) {
		if (argv[1][0]=='-') {
			if (strcmp(argv[1],"-D")==0) {
				initKeyTable(NULL);
				dumpKeys();
			} else useage(argv[0]);
			exit(0);
		}
		strcpy(paths[0],argv[1]);
	} else if (paths[0][0]=='\0') getcwd(paths[0],512+256);
	if (argc==3) strcpy(paths[1],argv[2]);
	else if (paths[1][0]=='\0') getcwd(paths[1],512+256);
#ifdef DEBUG
	debug = fopen("gc.debug","w");
#endif
	/* Make sure boths paths are accessible and absolute */
	chdir(paths[1]); getcwd(paths[1],512+256);
	chdir(paths[0]); getcwd(paths[0],512+256);
	initKeyTable(evar);
	setupCurses(1);
	readList(0,NULL,1);
	readList(1,NULL,1);
	tempName = tmpnam(NULL);
	i_paint();
	loop();
}

