/*********************************************************
	gc.h - some configureable stuff at the start...
**********************************************************/

/*
 * Specify your favourite default file paginator and editor here.
 * These only work with ANSI C compilers. Otherwise you will have
 * to change the code in gc.h explicitly to modify the defaults,
 * which are /bin/pg and /bin/vi.
 */

#define PAGER	"/bin/less"
#define EDITOR	"/bin/vi"

/* If you get a message about mode_t being undefined, uncomment the
	next line. */

/* typedef unsigned short mode_t; */

/*************************************************/
/* NOT MUCH SHOULD HAVE TO BE CHANGED BELOW HERE */
/*************************************************/

#ifndef MAXFILENAME
#define MAXFILENAME	35	/* Maximum displayed filename length */
#endif

#ifndef MAXARGS
#define MAXARGS		512	/* Maximum number of file names per window */
#endif

/*
 * Curses implementation specifics
 */

#ifdef svr4
#define UPDATE		wnoutrefresh
#define REFRESH		doupdate
#define KEYPAD(w,f)	keypad(w,f)
#define BEEP		beep()
#define WGETNSTR(w,b,l)		wgetnstr(w,b,l)
#else
#define UPDATE		wrefresh
#define REFRESH		refresh
#define KEYPAD(w,f) 	/* Not supported */
#define BEEP		putc(7,stdout)
#define WGETNSTR(w,b,l)	wgetstr(w,b)
#endif

/*
 * Lots of curses key defs, for those whose curses doesn't define them.
 */

#ifndef KEY_DOWN
#define KEY_DOWN 0402	/* Sent by terminal left arrow key */
#endif

#ifndef KEY_UP
#define KEY_UP 0403	/* Sent by terminal left arrow key */
#endif

#ifndef KEY_LEFT
#define KEY_LEFT 0404	/* Sent by terminal left arrow key */
#endif

#ifndef KEY_RIGHT
#define KEY_RIGHT 0405	/* Sent by terminal right arrow key */
#endif

#ifndef KEY_PPAGE
#define KEY_PPAGE 0523
#endif

#ifndef KEY_NPAGE
#define KEY_NPAGE 0522
#endif

#ifndef KEY_SF
#define KEY_SF 0520
#endif

#ifndef KEY_SR
#define KEY_SR 0521
#endif

#ifndef KEY_HOME
#define KEY_HOME	0406
#endif

#ifndef KEY_END
#define KEY_END	0550
#endif

#ifndef KEY_HELP
#define KEY_HELP 0553
#endif

#ifndef KEY_F0
#define KEY_F0 0410	/* Sent by terminal left arrow key */
#endif

#ifndef KEY_F
#define KEY_F(n) (KEY_F0 + (n))
#endif

/* 
 * POSIX file attribute macros
 */

#ifndef S_ISDIR				/* macro to test for directory file */
#define	S_ISDIR( mode )		(((mode) & S_IFMT) == S_IFDIR)
#endif

#ifndef S_ISCHR				/* macro to test for character special file */
#define	S_ISCHR( mode )		(((mode) & S_IFMT) == S_IFCHR)
#endif

#ifndef S_ISBLK				/* macro to test for block special file */
#define	S_ISBLK( mode )		(((mode) & S_IFMT) == S_IFBLK)
#endif

#ifndef S_ISFIFO			/* macro to test for fifo */
#define	S_ISFIFO( mode )	(((mode) & S_IFMT) == S_IFIFO)
#endif

#ifndef S_ISREG				/* macro to test for regular file */
#define	S_ISREG( mode )		(((mode) & S_IFMT) == S_IFREG)
#endif

#ifndef S_ISLNK
#define S_ISLNK(m)	0
#endif

/***************************************************************/

#define HEIGHT		LINES
#define NUM_INFO_LNS	4
#define NUM_FILES	(HEIGHT-NUM_INFO_LNS-5)

typedef struct {
	char		*name;
	unsigned char	flag;
	char		typID;
	off_t		size;
	mode_t		mode;
	uid_t		uid;
	gid_t		gid;
	time_t		modtime;
	time_t		acctime;
} fInfo_t;

#define MAX_CMDS	256	

#define SETSIZE	((MAX_CMDS+31)/32)

typedef unsigned long set_t[SETSIZE];

/*
 * Internal Commands
 */

#define CMD_PAINT	0 	/* Repaint the screen			*/
#define CMD_QUIT	1 	/* Leave gc				*/
#define CMD_HELP	2 	/* Show help screens			*/
#define CMD_CMDMODE	3 	/* Enter command mode			*/
#define CMD_READ	4 	/* Reread paths				*/
#define CMD_WAIT	5 	/* Prompt for and wait for a keypress	*/
#define	CMD_SRCH	6 	/* Start incremental filename search	*/
#define CMD_CLR		7 	/* Clear the screen			*/
#define	CMD_SWAP	8 	/* Swap windows				*/
#define	CMD_UP		9 	/* Up one line				*/
#define	CMD_DOWN	10	/* Down one line			*/
#define	CMD_PGUP	11	/* Page up				*/
#define	CMD_PGDN	12	/* Page down				*/
#define	CMD_HOME	13	/* Cursor Home				*/
#define	CMD_END		14	/* Cursor End				*/
#define	CMD_TOGGLE	15	/* Toggle current file selection state	*/
#define	CMD_INVERT	16	/* Toggle all selections		*/
#define CMD_BEEP	17	/* Beep terminal			*/
#define CMD_CD		18	/* Change directory			*/
#define CMD_CMDP	19 	/* Enter command mode			*/

static void i_paint(), i_quit(), i_help(), i_cmd(), i_cmdp(), i_read(),
	i_wait(), i_search(), i_clr(), i_swap(), i_up(), i_down(),
	i_pgup(), i_pgdn(), i_home(), i_end(), i_inv(),
	i_invall(), i_beep(), i_cd(), clearCmdWin(), showPath(),
	normalise();

#if	__STDC__
typedef void (*internalCmd_t)(void);
#else
typedef void (*internalCmd_t)();
#endif

/*
 * command flags
 */

#define C_CLR	1
#define C_CLRM	2
#define C_CATCH	4
#define C_MULTI	8

/*
 * command file type restriction flags
 */

#define F_REG	1
#define F_EXEC	2
#define F_DIR	4
#define F_PIPE	8
#define F_CDEV	16
#define F_BDEV	32
#define F_LINK	64
#define F_ANY	127

typedef struct {
	char	flags,
		restricts;
	char	*regexp;
	char	*actions;
} cmdInfo_t;

typedef struct {
	set_t	cmds;
	char	*help;
	char	multi,
		pre,
		post;
} keyInfo_t;


